"""Library specific errors."""


class PreviewClientLimitError(RuntimeError):
    """Limited usage of a preview client exceeded."""


class ClientResolutionError(RuntimeError):
    """Failed resolution of an unspecified client."""



import time
from abc import ABC, abstractmethod


class RetryStrategy(ABC):
    """Abstract base class that defines the interface for any retry strategy.

    See :obj:`MaxRetries`, :obj:`MaxWaitTime`, and :obj:`RegularPeriodRetry` for
    example implementations.
    """

    @abstractmethod
    def is_done(self) -> bool:
        """Check if the retry strategy has exhausted its retries."""
        pass

    @abstractmethod
    def wait(self) -> None:
        """Delay implementation before the next retry attempt."""
        pass

    @abstractmethod
    def step(self) -> None:
        """Update internal attributes of the retry strategy."""
        pass


class MaxRetries(RetryStrategy):
    """Check for submissions status every 100 ms and retry a maximum of
    `max_retries` times.

    Parameters
    ----------
    max_retries : int
        Max number of retries.
    """

    def __init__(self, max_retries: int = 20):
        if max_retries < 1:
            raise ValueError("max_retries must be at least 1.")
        self.n_retries = 0
        self.max_retries = max_retries

    def step(self):
        """Increment the number of retries by one."""
        self.n_retries += 1

    def wait(self):
        """Wait for 0.1 seconds between retries."""
        time.sleep(0.1)

    def is_done(self) -> bool:
        """Check if the number of retries is bigger or equal to specified
        maximum number of retries."""
        return self.n_retries >= self.max_retries


class MaxWaitTime(RetryStrategy):
    """Check for submissions status every 100 ms and wait for all submissions
    a maximum of `max_wait_time` (seconds).

    Parameters
    ----------
    max_wait_time_sec : float
        Maximum waiting time (in seconds).
    """

    def __init__(self, max_wait_time_sec: float = 5 * 60):
        self.max_wait_time_sec = max_wait_time_sec
        self.total_wait_time = 0

    def step(self):
        """Add 0.1 seconds to total waiting time."""
        self.total_wait_time += 0.1

    def wait(self):
        """Wait (sleep) for 0.1 seconds."""
        time.sleep(0.1)

    def is_done(self):
        """Check if the total waiting time is bigger or equal to the specified
        maximum waiting time."""
        return self.total_wait_time >= self.max_wait_time_sec


class RegularPeriodRetry(RetryStrategy):
    """Check for submissions status periodically for indefinite amount of time.

    Parameters
    ----------
    wait_time_sec : float
        Wait time between retries (in seconds).
    """

    def __init__(self, wait_time_sec: float = 0.1):
        self.wait_time_sec = wait_time_sec

    def wait(self):
        """Wait for `wait_time_sec` seconds."""
        time.sleep(self.wait_time_sec)

    def is_done(self) -> bool:
        """Return False, as this retry strategy is indefinite."""
        return False

    def step(self) -> None:
        """Satisfy the interface with a dummy implementation."""
        pass



from typing import ClassVar, Optional, Union

import requests

from .base_types import Config, Iterable, Language, LanguageAlias
from .data import LANGUAGE_TO_LANGUAGE_ID
from .retry import RetryStrategy
from .submission import Submission, Submissions
from .utils import handle_too_many_requests_error_for_preview_client


class Client:
    """Base class for all clients.

    Parameters
    ----------
    endpoint : str
        Client's default endpoint.
    auth_headers : dict
        Request authentication headers.

    Attributes
    ----------
    API_KEY_ENV : str
        Environment variable where judge0-python should look for API key for
        the client. Set to default values for RapidAPI and ATD clients.
    """

    # Environment variable where judge0-python should look for API key for
    # the client. Set to default values for RapidAPI and ATD clients.
    API_KEY_ENV: ClassVar[str] = None

    def __init__(
        self,
        endpoint,
        auth_headers,
        *,
        retry_strategy: Optional[RetryStrategy] = None,
    ) -> None:
        self.endpoint = endpoint
        self.auth_headers = auth_headers
        self.retry_strategy = retry_strategy
        self.session = requests.Session()

        try:
            self.languages = self.get_languages()
            self.config = self.get_config_info()
        except Exception as e:
            home_url = getattr(self, "HOME_URL", None)
            raise RuntimeError(
                f"Authentication failed. Visit {home_url} to get or "
                "review your authentication credentials."
            ) from e

    def __del__(self):
        self.session.close()

    @handle_too_many_requests_error_for_preview_client
    def get_about(self) -> dict:
        """Get general information about judge0.

        Returns
        -------
        dict
            General information about judge0.
        """
        response = self.session.get(
            f"{self.endpoint}/about",
            headers=self.auth_headers,
        )
        response.raise_for_status()
        return response.json()

    @handle_too_many_requests_error_for_preview_client
    def get_config_info(self) -> Config:
        """Get information about client's configuration.

        Returns
        -------
        Config
            Client's configuration.
        """
        response = self.session.get(
            f"{self.endpoint}/config_info",
            headers=self.auth_headers,
        )
        response.raise_for_status()
        return Config(**response.json())

    @handle_too_many_requests_error_for_preview_client
    def get_language(self, language_id: int) -> Language:
        """Get language corresponding to the id.

        Parameters
        ----------
        language_id : int
            Language id.

        Returns
        -------
        Language
            Language corresponding to the passed id.
        """
        request_url = f"{self.endpoint}/languages/{language_id}"
        response = self.session.get(request_url, headers=self.auth_headers)
        response.raise_for_status()
        return Language(**response.json())

    @handle_too_many_requests_error_for_preview_client
    def get_languages(self) -> list[Language]:
        """Get a list of supported languages.

        Returns
        -------
        list of language
            A list of supported languages.
        """
        request_url = f"{self.endpoint}/languages"
        response = self.session.get(request_url, headers=self.auth_headers)
        response.raise_for_status()
        return [Language(**lang_dict) for lang_dict in response.json()]

    @handle_too_many_requests_error_for_preview_client
    def get_statuses(self) -> list[dict]:
        """Get a list of possible submission statuses.

        Returns
        -------
        list of dict
            A list of possible submission statues.
        """
        response = self.session.get(
            f"{self.endpoint}/statuses",
            headers=self.auth_headers,
        )
        response.raise_for_status()
        return response.json()

    @property
    def version(self):
        """Property corresponding to the current client's version."""
        if not hasattr(self, "_version"):
            _version = self.get_about()["version"]
            setattr(self, "_version", _version)
        return self._version

    def get_language_id(self, language: Union[LanguageAlias, int]) -> int:
        """Get language id corresponding to the language alias for the client.

        Parameters
        ----------
        language : LanguageAlias or int
            Language alias or language id.

        Returns
        -------
            Language id corresponding to the language alias.
        """
        if isinstance(language, LanguageAlias):
            supported_language_ids = LANGUAGE_TO_LANGUAGE_ID[self.version]
            language = supported_language_ids.get(language, -1)
        return language

    def is_language_supported(self, language: Union[LanguageAlias, int]) -> bool:
        """Check if language is supported by the client.

        Parameters
        ----------
        language : LanguageAlias or int
            Language alias or language id.

        Returns
        -------
        bool
            Return True if language is supported by the client, otherwise returns
            False.
        """
        language_id = self.get_language_id(language)
        return any(language_id == lang.id for lang in self.languages)

    @handle_too_many_requests_error_for_preview_client
    def create_submission(self, submission: Submission) -> Submission:
        """Send submission for execution to a client.

        Directly send a submission to create_submission route for execution.

        Parameters
        ----------
        submission : Submission
            A submission to create.

        Returns
        -------
        Submission
            A submission with updated token attribute.
        """
        # Check if the client supports the language specified in the submission.
        if not self.is_language_supported(language=submission.language):
            raise RuntimeError(
                f"Client {type(self).__name__} does not support language with "
                f"id {submission.language}!"
            )

        params = {
            "base64_encoded": "true",
            "wait": "false",
        }

        body = submission.as_body(self)

        response = self.session.post(
            f"{self.endpoint}/submissions",
            json=body,
            params=params,
            headers=self.auth_headers,
        )
        response.raise_for_status()

        submission.set_attributes(response.json())

        return submission

    @handle_too_many_requests_error_for_preview_client
    def get_submission(
        self,
        submission: Submission,
        *,
        fields: Optional[Union[str, Iterable[str]]] = None,
    ) -> Submission:
        """Get submissions status.

        Directly send submission's token to get_submission route for status
        check. By default, all submissions attributes (fields) are requested.

        Parameters
        ----------
        submission : Submission
            Submission to update.

        Returns
        -------
        Submission
            A Submission with updated attributes.
        """
        params = {
            "base64_encoded": "true",
        }

        if isinstance(fields, str):
            fields = [fields]

        if fields is not None:
            params["fields"] = ",".join(fields)
        else:
            params["fields"] = "*"

        response = self.session.get(
            f"{self.endpoint}/submissions/{submission.token}",
            params=params,
            headers=self.auth_headers,
        )
        response.raise_for_status()

        submission.set_attributes(response.json())

        return submission

    @handle_too_many_requests_error_for_preview_client
    def create_submissions(self, submissions: Submissions) -> Submissions:
        """Send submissions for execution to a client.

        Directly send submissions to create_submissions route for execution.
        Cannot handle more submissions than the client supports.

        Parameters
        ----------
        submissions : Submissions
            A sequence of submissions to create.

        Returns
        -------
        Submissions
            A sequence of submissions with updated token attribute.
        """
        for submission in submissions:
            if not self.is_language_supported(language=submission.language):
                raise RuntimeError(
                    f"Client {type(self).__name__} does not support language "
                    f"{submission.language}!"
                )

        submissions_body = [submission.as_body(self) for submission in submissions]

        response = self.session.post(
            f"{self.endpoint}/submissions/batch",
            headers=self.auth_headers,
            params={"base64_encoded": "true"},
            json={"submissions": submissions_body},
        )
        response.raise_for_status()

        for submission, attrs in zip(submissions, response.json()):
            submission.set_attributes(attrs)

        return submissions

    @handle_too_many_requests_error_for_preview_client
    def get_submissions(
        self,
        submissions: Submissions,
        *,
        fields: Optional[Union[str, Iterable[str]]] = None,
    ) -> Submissions:
        """Get submissions status.

        Directly send submissions' tokens to get_submissions route for status
        check. By default, all submissions attributes (fields) are requested.
        Cannot handle more submissions than the client supports.

        Parameters
        ----------
        submissions : Submissions
            Submissions to update.

        Returns
        -------
        Submissions
            A sequence of submissions with updated attributes.
        """
        params = {
            "base64_encoded": "true",
        }

        if isinstance(fields, str):
            fields = [fields]

        if fields is not None:
            params["fields"] = ",".join(fields)
        else:
            params["fields"] = "*"

        tokens = ",".join([submission.token for submission in submissions])
        params["tokens"] = tokens

        response = self.session.get(
            f"{self.endpoint}/submissions/batch",
            params=params,
            headers=self.auth_headers,
        )
        response.raise_for_status()

        for submission, attrs in zip(submissions, response.json()["submissions"]):
            submission.set_attributes(attrs)

        return submissions


class ATD(Client):
    """Base class for all AllThingsDev clients.

    Parameters
    ----------
    endpoint : str
        Default request endpoint.
    host_header_value : str
        Value for the x-apihub-host header.
    api_key : str
        AllThingsDev API key.
    **kwargs : dict
        Additional keyword arguments for the base Client.
    """

    API_KEY_ENV: ClassVar[str] = "JUDGE0_ATD_API_KEY"

    def __init__(self, endpoint, host_header_value, api_key, **kwargs):
        self.api_key = api_key
        super().__init__(
            endpoint,
            {
                "x-apihub-host": host_header_value,
                "x-apihub-key": api_key,
            },
            **kwargs,
        )

    def _update_endpoint_header(self, header_value):
        self.auth_headers["x-apihub-endpoint"] = header_value


class ATDJudge0CE(ATD):
    """AllThingsDev client for CE flavor.

    Parameters
    ----------
    api_key : str
        AllThingsDev API key.
    **kwargs : dict
        Additional keyword arguments for the base Client.
    """

    DEFAULT_ENDPOINT: ClassVar[str] = (
        "https://judge0-ce.proxy-production.allthingsdev.co"
    )
    DEFAULT_HOST: ClassVar[str] = "Judge0-CE.allthingsdev.co"
    HOME_URL: ClassVar[str] = (
        "https://www.allthingsdev.co/apimarketplace/judge0-ce/66b683c8b7b7ad054eb6ff8f"
    )

    DEFAULT_ABOUT_ENDPOINT: ClassVar[str] = "01fc1c98-ceee-4f49-8614-f2214703e25f"
    DEFAULT_CONFIG_INFO_ENDPOINT: ClassVar[str] = "b7aab45d-5eb0-4519-b092-89e5af4fc4f3"
    DEFAULT_LANGUAGE_ENDPOINT: ClassVar[str] = "a50ae6b1-23c1-40eb-b34c-88bc8cf2c764"
    DEFAULT_LANGUAGES_ENDPOINT: ClassVar[str] = "03824deb-bd18-4456-8849-69d78e1383cc"
    DEFAULT_STATUSES_ENDPOINT: ClassVar[str] = "c37b603f-6f99-4e31-a361-7154c734f19b"
    DEFAULT_CREATE_SUBMISSION_ENDPOINT: ClassVar[str] = (
        "6e65686d-40b0-4bf7-a12f-1f6d033c4473"
    )
    DEFAULT_GET_SUBMISSION_ENDPOINT: ClassVar[str] = (
        "b7032b8b-86da-40b4-b9d3-b1f5e2b4ee1e"
    )
    DEFAULT_CREATE_SUBMISSIONS_ENDPOINT: ClassVar[str] = (
        "402b857c-1126-4450-bfd8-22e1f2cbff2f"
    )
    DEFAULT_GET_SUBMISSIONS_ENDPOINT: ClassVar[str] = (
        "e42f2a26-5b02-472a-80c9-61c4bdae32ec"
    )

    def __init__(self, api_key, **kwargs):
        super().__init__(
            self.DEFAULT_ENDPOINT,
            self.DEFAULT_HOST,
            api_key,
            **kwargs,
        )

    def get_about(self) -> dict:
        self._update_endpoint_header(self.DEFAULT_ABOUT_ENDPOINT)
        return super().get_about()

    def get_config_info(self) -> Config:
        self._update_endpoint_header(self.DEFAULT_CONFIG_INFO_ENDPOINT)
        return super().get_config_info()

    def get_language(self, language_id) -> Language:
        self._update_endpoint_header(self.DEFAULT_LANGUAGE_ENDPOINT)
        return super().get_language(language_id)

    def get_languages(self) -> list[Language]:
        self._update_endpoint_header(self.DEFAULT_LANGUAGES_ENDPOINT)
        return super().get_languages()

    def get_statuses(self) -> list[dict]:
        self._update_endpoint_header(self.DEFAULT_STATUSES_ENDPOINT)
        return super().get_statuses()

    def create_submission(self, submission: Submission) -> Submission:
        self._update_endpoint_header(self.DEFAULT_CREATE_SUBMISSION_ENDPOINT)
        return super().create_submission(submission)

    def get_submission(
        self,
        submission: Submission,
        *,
        fields: Optional[Union[str, Iterable[str]]] = None,
    ) -> Submission:
        self._update_endpoint_header(self.DEFAULT_GET_SUBMISSION_ENDPOINT)
        return super().get_submission(submission, fields=fields)

    def create_submissions(self, submissions: Submissions) -> Submissions:
        self._update_endpoint_header(self.DEFAULT_CREATE_SUBMISSIONS_ENDPOINT)
        return super().create_submissions(submissions)

    def get_submissions(
        self,
        submissions: Submissions,
        *,
        fields: Optional[Union[str, Iterable[str]]] = None,
    ) -> Submissions:
        self._update_endpoint_header(self.DEFAULT_GET_SUBMISSIONS_ENDPOINT)
        return super().get_submissions(submissions, fields=fields)


class ATDJudge0ExtraCE(ATD):
    """AllThingsDev client for Extra CE flavor.

    Parameters
    ----------
    api_key : str
        AllThingsDev API key.
    **kwargs : dict
        Additional keyword arguments for the base Client.
    """

    DEFAULT_ENDPOINT: ClassVar[str] = (
        "https://judge0-extra-ce.proxy-production.allthingsdev.co"
    )
    DEFAULT_HOST: ClassVar[str] = "Judge0-Extra-CE.allthingsdev.co"
    HOME_URL: ClassVar[str] = (
        "https://www.allthingsdev.co/apimarketplace/judge0-extra-ce/"
        "66b68838b7b7ad054eb70690"
    )

    DEFAULT_ABOUT_ENDPOINT: ClassVar[str] = "1fd631a1-be6a-47d6-bf4c-987e357e3096"
    DEFAULT_CONFIG_INFO_ENDPOINT: ClassVar[str] = "46e05354-2a43-436a-9458-5d111456f0ff"
    DEFAULT_LANGUAGE_ENDPOINT: ClassVar[str] = "10465a84-2a2c-4213-845f-45e3c04a5867"
    DEFAULT_LANGUAGES_ENDPOINT: ClassVar[str] = "774ecece-1200-41f7-a992-38f186c90803"
    DEFAULT_STATUSES_ENDPOINT: ClassVar[str] = "a2843b3c-673d-4966-9a14-2e7d76dcd0cb"
    DEFAULT_CREATE_SUBMISSION_ENDPOINT: ClassVar[str] = (
        "be2d195e-dd58-4770-9f3c-d6c0fbc2b6e5"
    )
    DEFAULT_GET_SUBMISSION_ENDPOINT: ClassVar[str] = (
        "c3a457cd-37a6-4106-97a8-9e60a223abbc"
    )
    DEFAULT_CREATE_SUBMISSIONS_ENDPOINT: ClassVar[str] = (
        "c64df5d3-edfd-4b08-8687-561af2f80d2f"
    )
    DEFAULT_GET_SUBMISSIONS_ENDPOINT: ClassVar[str] = (
        "5d173718-8e6a-4cf5-9d8c-db5e6386d037"
    )

    def __init__(self, api_key, **kwargs):
        super().__init__(
            self.DEFAULT_ENDPOINT,
            self.DEFAULT_HOST,
            api_key,
            **kwargs,
        )

    def get_about(self) -> dict:
        self._update_endpoint_header(self.DEFAULT_ABOUT_ENDPOINT)
        return super().get_about()

    def get_config_info(self) -> Config:
        self._update_endpoint_header(self.DEFAULT_CONFIG_INFO_ENDPOINT)
        return super().get_config_info()

    def get_language(self, language_id) -> Language:
        self._update_endpoint_header(self.DEFAULT_LANGUAGE_ENDPOINT)
        return super().get_language(language_id)

    def get_languages(self) -> list[Language]:
        self._update_endpoint_header(self.DEFAULT_LANGUAGES_ENDPOINT)
        return super().get_languages()

    def get_statuses(self) -> list[dict]:
        self._update_endpoint_header(self.DEFAULT_STATUSES_ENDPOINT)
        return super().get_statuses()

    def create_submission(self, submission: Submission) -> Submission:
        self._update_endpoint_header(self.DEFAULT_CREATE_SUBMISSION_ENDPOINT)
        return super().create_submission(submission)

    def get_submission(
        self,
        submission: Submission,
        *,
        fields: Optional[Union[str, Iterable[str]]] = None,
    ) -> Submission:
        self._update_endpoint_header(self.DEFAULT_GET_SUBMISSION_ENDPOINT)
        return super().get_submission(submission, fields=fields)

    def create_submissions(self, submissions: Submissions) -> Submissions:
        self._update_endpoint_header(self.DEFAULT_CREATE_SUBMISSIONS_ENDPOINT)
        return super().create_submissions(submissions)

    def get_submissions(
        self,
        submissions: Submissions,
        *,
        fields: Optional[Union[str, Iterable[str]]] = None,
    ) -> Submissions:
        self._update_endpoint_header(self.DEFAULT_GET_SUBMISSIONS_ENDPOINT)
        return super().get_submissions(submissions, fields=fields)


class Rapid(Client):
    """Base class for all RapidAPI clients.

    Parameters
    ----------
    endpoint : str
        Default request endpoint.
    host_header_value : str
        Value for the x-rapidapi-host header.
    api_key : str
        RapidAPI API key.
    **kwargs : dict
        Additional keyword arguments for the base Client.
    """

    API_KEY_ENV: ClassVar[str] = "JUDGE0_RAPID_API_KEY"

    def __init__(self, endpoint, host_header_value, api_key, **kwargs):
        self.api_key = api_key
        super().__init__(
            endpoint,
            {
                "x-rapidapi-host": host_header_value,
                "x-rapidapi-key": api_key,
            },
            **kwargs,
        )


class RapidJudge0CE(Rapid):
    """RapidAPI client for CE flavor.

    Parameters
    ----------
    api_key : str
        RapidAPI API key.
    **kwargs : dict
        Additional keyword arguments for the base Client.
    """

    DEFAULT_ENDPOINT: ClassVar[str] = "https://judge0-ce.p.rapidapi.com"
    DEFAULT_HOST: ClassVar[str] = "judge0-ce.p.rapidapi.com"
    HOME_URL: ClassVar[str] = "https://rapidapi.com/judge0-official/api/judge0-ce"

    def __init__(self, api_key, **kwargs):
        super().__init__(
            self.DEFAULT_ENDPOINT,
            self.DEFAULT_HOST,
            api_key,
            **kwargs,
        )


class RapidJudge0ExtraCE(Rapid):
    """RapidAPI client for Extra CE flavor.

    Parameters
    ----------
    api_key : str
        RapidAPI API key.
    **kwargs : dict
        Additional keyword arguments for the base Client.
    """

    DEFAULT_ENDPOINT: ClassVar[str] = "https://judge0-extra-ce.p.rapidapi.com"
    DEFAULT_HOST: ClassVar[str] = "judge0-extra-ce.p.rapidapi.com"
    HOME_URL: ClassVar[str] = "https://rapidapi.com/judge0-official/api/judge0-extra-ce"

    def __init__(self, api_key, **kwargs):
        super().__init__(
            self.DEFAULT_ENDPOINT,
            self.DEFAULT_HOST,
            api_key,
            **kwargs,
        )


class Judge0Cloud(Client):
    """Base class for all Judge0 Cloud clients.

    Parameters
    ----------
    endpoint : str
        Default request endpoint.
    auth_headers : str or dict
        Judge0 Cloud authentication headers, either as a JSON string or a dictionary.
    **kwargs : dict
        Additional keyword arguments for the base Client.
    """

    def __init__(self, endpoint, auth_headers=None, **kwargs):
        if isinstance(auth_headers, str):
            from json import loads

            auth_headers = loads(auth_headers)

        super().__init__(
            endpoint,
            auth_headers,
            **kwargs,
        )


class Judge0CloudCE(Judge0Cloud):
    """Judge0 Cloud client for CE flavor.

    Parameters
    ----------
    endpoint : str
        Default request endpoint.
    auth_headers : str or dict
        Judge0 Cloud authentication headers, either as a JSON string or a dictionary.
    **kwargs : dict
        Additional keyword arguments for the base Client.
    """

    DEFAULT_ENDPOINT: ClassVar[str] = "https://ce.judge0.com"
    HOME_URL: ClassVar[str] = "https://ce.judge0.com"
    API_KEY_ENV: ClassVar[str] = "JUDGE0_CLOUD_CE_AUTH_HEADERS"

    def __init__(self, auth_headers=None, **kwargs):
        super().__init__(
            self.DEFAULT_ENDPOINT,
            auth_headers,
            **kwargs,
        )


class Judge0CloudExtraCE(Judge0Cloud):
    """Judge0 Cloud client for Extra CE flavor.

    Parameters
    ----------
    endpoint : str
        Default request endpoint.
    auth_headers : str or dict
        Judge0 Cloud authentication headers, either as a JSON string or a dictionary.
    **kwargs : dict
        Additional keyword arguments for the base Client.
    """

    DEFAULT_ENDPOINT: ClassVar[str] = "https://extra-ce.judge0.com"
    HOME_URL: ClassVar[str] = "https://extra-ce.judge0.com"
    API_KEY_ENV: ClassVar[str] = "JUDGE0_CLOUD_EXTRA_CE_AUTH_HEADERS"

    def __init__(self, auth_headers=None, **kwargs):
        super().__init__(self.DEFAULT_ENDPOINT, auth_headers, **kwargs)


CE = (Judge0CloudCE, RapidJudge0CE, ATDJudge0CE)
EXTRA_CE = (Judge0CloudExtraCE, RapidJudge0ExtraCE, ATDJudge0ExtraCE)



import logging
import os

from typing import Union

from .api import (
    async_execute,
    async_run,
    execute,
    get_client,
    run,
    sync_execute,
    sync_run,
    wait,
)
from .base_types import Flavor, Language, LanguageAlias, Status, TestCase
from .clients import (
    ATD,
    ATDJudge0CE,
    ATDJudge0ExtraCE,
    Client,
    Judge0Cloud,
    Judge0CloudCE,
    Judge0CloudExtraCE,
    Rapid,
    RapidJudge0CE,
    RapidJudge0ExtraCE,
)
from .filesystem import File, Filesystem
from .retry import MaxRetries, MaxWaitTime, RegularPeriodRetry
from .submission import Submission

__version__ = "0.0.5"

__all__ = [
    "ATD",
    "ATDJudge0CE",
    "ATDJudge0ExtraCE",
    "Client",
    "File",
    "Filesystem",
    "Judge0Cloud",
    "Judge0CloudCE",
    "Judge0CloudExtraCE",
    "Language",
    "LanguageAlias",
    "MaxRetries",
    "MaxWaitTime",
    "Rapid",
    "RapidJudge0CE",
    "RapidJudge0ExtraCE",
    "RegularPeriodRetry",
    "Status",
    "Submission",
    "TestCase",
    "async_execute",
    "async_run",
    "execute",
    "get_client",
    "run",
    "sync_execute",
    "sync_run",
    "wait",
]

JUDGE0_IMPLICIT_CE_CLIENT = None
JUDGE0_IMPLICIT_EXTRA_CE_CLIENT = None

logger = logging.getLogger(__name__)


def _get_implicit_client(flavor: Flavor) -> Client:
    global JUDGE0_IMPLICIT_CE_CLIENT, JUDGE0_IMPLICIT_EXTRA_CE_CLIENT

    # Implicit clients are already set.
    if flavor == Flavor.CE and JUDGE0_IMPLICIT_CE_CLIENT is not None:
        return JUDGE0_IMPLICIT_CE_CLIENT
    if flavor == Flavor.EXTRA_CE and JUDGE0_IMPLICIT_EXTRA_CE_CLIENT is not None:
        return JUDGE0_IMPLICIT_EXTRA_CE_CLIENT

    try:
        from dotenv import load_dotenv

        load_dotenv()
    except:  # noqa: E722
        pass

    # Let's check if we can find a self-hosted client.
    client = _get_custom_client(flavor)

    # Try to find one of the API keys for hub clients.
    if client is None:
        client = _get_hub_client(flavor)

    # If we didn't find any of the possible keys, initialize
    # the preview client based on the flavor.
    if client is None:
        client = _get_preview_client(flavor)

    if flavor == Flavor.CE:
        JUDGE0_IMPLICIT_CE_CLIENT = client
    else:
        JUDGE0_IMPLICIT_EXTRA_CE_CLIENT = client

    return client


def _get_preview_client(flavor: Flavor) -> Union[Judge0CloudCE, Judge0CloudExtraCE]:
    logger.warning(
        "You are using a preview version of the client which is not recommended"
        " for production.\n"
        "For production, please specify your API key in the environment variable."
    )

    if flavor == Flavor.CE:
        return Judge0CloudCE()
    else:
        return Judge0CloudExtraCE()


def _get_custom_client(flavor: Flavor) -> Union[Client, None]:
    from json import loads

    ce_endpoint = os.getenv("JUDGE0_CE_ENDPOINT")
    ce_auth_header = os.getenv("JUDGE0_CE_AUTH_HEADERS")
    extra_ce_endpoint = os.getenv("JUDGE0_EXTRA_CE_ENDPOINT")
    extra_ce_auth_header = os.getenv("JUDGE0_EXTRA_CE_AUTH_HEADERS")

    if flavor == Flavor.CE and ce_endpoint is not None and ce_auth_header is not None:
        return Client(
            endpoint=ce_endpoint,
            auth_headers=loads(ce_auth_header),
        )

    if (
        flavor == Flavor.EXTRA_CE
        and extra_ce_endpoint is not None
        and extra_ce_auth_header is not None
    ):
        return Client(
            endpoint=extra_ce_endpoint,
            auth_headers=loads(extra_ce_auth_header),
        )

    return None


def _get_hub_client(flavor: Flavor) -> Union[Client, None]:
    from .clients import CE, EXTRA_CE

    if flavor == Flavor.CE:
        client_classes = CE
    else:
        client_classes = EXTRA_CE

    for client_class in client_classes:
        api_key = os.getenv(client_class.API_KEY_ENV)
        if api_key is not None:
            client = client_class(api_key)
            break
    else:
        client = None

    return client


CE = Flavor.CE
EXTRA_CE = Flavor.EXTRA_CE

ASSEMBLY = LanguageAlias.ASSEMBLY
BASH = LanguageAlias.BASH
BASIC = LanguageAlias.BASIC
BOSQUE = LanguageAlias.BOSQUE
C = LanguageAlias.C
C3 = LanguageAlias.C3
CLOJURE = LanguageAlias.CLOJURE
COBOL = LanguageAlias.COBOL
COMMON_LISP = LanguageAlias.COMMON_LISP
CPP = LanguageAlias.CPP
CPP_CLANG = LanguageAlias.CPP_CLANG
CPP_GCC = LanguageAlias.CPP_GCC
CPP_TEST = LanguageAlias.CPP_TEST
CPP_TEST_CLANG = LanguageAlias.CPP_TEST_CLANG
CPP_TEST_GCC = LanguageAlias.CPP_TEST_GCC
CSHARP = LanguageAlias.CSHARP
CSHARP_DOTNET = LanguageAlias.CSHARP_DOTNET
CSHARP_MONO = LanguageAlias.CSHARP_MONO
CSHARP_TEST = LanguageAlias.CSHARP_TEST
C_CLANG = LanguageAlias.C_CLANG
C_GCC = LanguageAlias.C_GCC
D = LanguageAlias.D
DART = LanguageAlias.DART
ELIXIR = LanguageAlias.ELIXIR
ERLANG = LanguageAlias.ERLANG
EXECUTABLE = LanguageAlias.EXECUTABLE
FORTRAN = LanguageAlias.FORTRAN
FSHARP = LanguageAlias.FSHARP
GO = LanguageAlias.GO
GROOVY = LanguageAlias.GROOVY
HASKELL = LanguageAlias.HASKELL
JAVA = LanguageAlias.JAVA
JAVAFX = LanguageAlias.JAVAFX
JAVASCRIPT = LanguageAlias.JAVASCRIPT
JAVA_JDK = LanguageAlias.JAVA_JDK
JAVA_OPENJDK = LanguageAlias.JAVA_OPENJDK
JAVA_TEST = LanguageAlias.JAVA_TEST
KOTLIN = LanguageAlias.KOTLIN
LUA = LanguageAlias.LUA
MPI_C = LanguageAlias.MPI_C
MPI_CPP = LanguageAlias.MPI_CPP
MPI_PYTHON = LanguageAlias.MPI_PYTHON
MULTI_FILE = LanguageAlias.MULTI_FILE
NIM = LanguageAlias.NIM
OBJECTIVE_C = LanguageAlias.OBJECTIVE_C
OCAML = LanguageAlias.OCAML
OCTAVE = LanguageAlias.OCTAVE
PASCAL = LanguageAlias.PASCAL
PERL = LanguageAlias.PERL
PHP = LanguageAlias.PHP
PLAIN_TEXT = LanguageAlias.PLAIN_TEXT
PROLOG = LanguageAlias.PROLOG
PYTHON = LanguageAlias.PYTHON
PYTHON2 = LanguageAlias.PYTHON2
PYTHON2_PYPY = LanguageAlias.PYTHON2_PYPY
PYTHON3 = LanguageAlias.PYTHON3
PYTHON3_PYPY = LanguageAlias.PYTHON3_PYPY
PYTHON_FOR_ML = LanguageAlias.PYTHON_FOR_ML
PYTHON_PYPY = LanguageAlias.PYTHON_PYPY
R = LanguageAlias.R
RUBY = LanguageAlias.RUBY
RUST = LanguageAlias.RUST
SCALA = LanguageAlias.SCALA
SQLITE = LanguageAlias.SQLITE
SWIFT = LanguageAlias.SWIFT
TYPESCRIPT = LanguageAlias.TYPESCRIPT
VISUAL_BASIC = LanguageAlias.VISUAL_BASIC



import copy

from dataclasses import dataclass
from enum import auto, IntEnum
from typing import Optional, Protocol, runtime_checkable, Sequence, Union

from pydantic import BaseModel

Iterable = Sequence

TestCaseType = Union["TestCase", list, tuple, dict]
TestCases = Iterable[TestCaseType]


@dataclass(frozen=True)
class TestCase:
    """Test case data model."""

    __test__ = False  # Needed to avoid pytest warning

    input: Optional[str] = None
    expected_output: Optional[str] = None

    @classmethod
    def from_record(
        cls, test_case: Union[TestCaseType, None]
    ) -> Union["TestCase", None]:
        """Create a TestCase from built-in types.

        Parameters
        ----------
        test_case: :obj:`TestCaseType` or None
            Test case data.

        Returns
        -------
        TestCase or None
            Created TestCase object or None if test_case is None.
        """
        if isinstance(test_case, (tuple, list)):
            test_case = {
                field: value
                for field, value in zip(("input", "expected_output"), test_case)
            }
        if isinstance(test_case, dict):
            return cls(
                input=test_case.get("input", None),
                expected_output=test_case.get("expected_output", None),
            )
        if isinstance(test_case, cls):
            return copy.deepcopy(test_case)
        if test_case is None:
            return None
        raise ValueError(
            f"Cannot create TestCase object from object of type {type(test_case)}."
        )


@runtime_checkable
class Encodable(Protocol):
    def encode(self) -> bytes:
        """Serialize the object to bytes."""
        ...


class Language(BaseModel):
    """Language data model.

    Stores information about a language supported by Judge0.
    """

    id: int
    name: str
    is_archived: Optional[bool] = None
    source_file: Optional[str] = None
    compile_cmd: Optional[str] = None
    run_cmd: Optional[str] = None


class LanguageAlias(IntEnum):
    """Language alias enumeration.

    Enumerates the programming languages supported by Judge0 client. Language
    alias is resolved to the latest version of the language supported by the
    selected client.
    """

    ASSEMBLY = auto()
    BASH = auto()
    BASIC = auto()
    BOSQUE = auto()
    C = auto()
    C3 = auto()
    CLOJURE = auto()
    COBOL = auto()
    COMMON_LISP = auto()
    CPP = auto()
    CPP_CLANG = auto()
    CPP_GCC = auto()
    CPP_TEST = auto()
    CPP_TEST_CLANG = auto()
    CPP_TEST_GCC = auto()
    CSHARP = auto()
    CSHARP_DOTNET = auto()
    CSHARP_MONO = auto()
    CSHARP_TEST = auto()
    C_CLANG = auto()
    C_GCC = auto()
    D = auto()
    DART = auto()
    ELIXIR = auto()
    ERLANG = auto()
    EXECUTABLE = auto()
    FORTRAN = auto()
    FSHARP = auto()
    GO = auto()
    GROOVY = auto()
    HASKELL = auto()
    JAVA = auto()
    JAVAFX = auto()
    JAVASCRIPT = auto()
    JAVA_JDK = auto()
    JAVA_OPENJDK = auto()
    JAVA_TEST = auto()
    KOTLIN = auto()
    LUA = auto()
    MPI_C = auto()
    MPI_CPP = auto()
    MPI_PYTHON = auto()
    MULTI_FILE = auto()
    NIM = auto()
    OBJECTIVE_C = auto()
    OCAML = auto()
    OCTAVE = auto()
    PASCAL = auto()
    PERL = auto()
    PHP = auto()
    PLAIN_TEXT = auto()
    PROLOG = auto()
    PYTHON = auto()
    PYTHON2 = auto()
    PYTHON2_PYPY = auto()
    PYTHON3 = auto()
    PYTHON3_PYPY = auto()
    PYTHON_FOR_ML = auto()
    PYTHON_PYPY = auto()
    R = auto()
    RUBY = auto()
    RUST = auto()
    SCALA = auto()
    SQLITE = auto()
    SWIFT = auto()
    TYPESCRIPT = auto()
    VISUAL_BASIC = auto()


class Flavor(IntEnum):
    """Flavor enumeration.

    Enumerates the flavors supported by Judge0 client.
    """

    CE = 0
    EXTRA_CE = 1


class Status(IntEnum):
    """Status enumeration.

    Enumerates possible status codes of a submission.
    """

    IN_QUEUE = 1
    PROCESSING = 2
    ACCEPTED = 3
    WRONG_ANSWER = 4
    TIME_LIMIT_EXCEEDED = 5
    COMPILATION_ERROR = 6
    RUNTIME_ERROR_SIGSEGV = 7
    RUNTIME_ERROR_SIGXFSZ = 8
    RUNTIME_ERROR_SIGFPE = 9
    RUNTIME_ERROR_SIGABRT = 10
    RUNTIME_ERROR_NZEC = 11
    RUNTIME_ERROR_OTHER = 12
    INTERNAL_ERROR = 13
    EXEC_FORMAT_ERROR = 14

    def __str__(self):
        return self.name.lower().replace("_", " ").title()


class Config(BaseModel):
    """Client config data model.

    Stores configuration data for the Judge0 client.
    """

    allow_enable_network: bool
    allow_enable_per_process_and_thread_memory_limit: bool
    allow_enable_per_process_and_thread_time_limit: bool
    allowed_languages_for_compile_options: list[str]
    callbacks_max_tries: int
    callbacks_timeout: float
    cpu_extra_time: float
    cpu_time_limit: float
    enable_additional_files: bool
    enable_batched_submissions: bool
    enable_callbacks: bool
    enable_command_line_arguments: bool
    enable_compiler_options: bool
    enable_network: bool
    enable_per_process_and_thread_memory_limit: bool
    enable_per_process_and_thread_time_limit: bool
    enable_submission_delete: bool
    enable_wait_result: bool
    maintenance_mode: bool
    max_cpu_extra_time: float
    max_cpu_time_limit: float
    max_extract_size: int
    max_file_size: int
    max_max_file_size: int
    max_max_processes_and_or_threads: int
    max_memory_limit: int
    max_number_of_runs: int
    max_processes_and_or_threads: int
    max_queue_size: int
    max_stack_limit: int
    max_submission_batch_size: int
    max_wall_time_limit: float
    memory_limit: int
    number_of_runs: int
    redirect_stderr_to_stdout: bool
    stack_limit: int
    submission_cache_duration: float
    use_docs_as_homepage: bool
    wall_time_limit: float



from .base_types import LanguageAlias

LANGUAGE_TO_LANGUAGE_ID = {
    "1.13.1": {
        LanguageAlias.ASSEMBLY: 45,
        LanguageAlias.BASH: 46,
        LanguageAlias.BASIC: 47,
        LanguageAlias.C: 50,
        LanguageAlias.CLOJURE: 86,
        LanguageAlias.COBOL: 77,
        LanguageAlias.COMMON_LISP: 55,
        LanguageAlias.CPP: 52,
        LanguageAlias.CPP_CLANG: 76,
        LanguageAlias.CPP_GCC: 52,
        LanguageAlias.CSHARP: 51,
        LanguageAlias.CSHARP_MONO: 51,
        LanguageAlias.C_CLANG: 75,
        LanguageAlias.C_GCC: 50,
        LanguageAlias.D: 56,
        LanguageAlias.ELIXIR: 57,
        LanguageAlias.ERLANG: 58,
        LanguageAlias.EXECUTABLE: 44,
        LanguageAlias.FORTRAN: 59,
        LanguageAlias.FSHARP: 87,
        LanguageAlias.GO: 60,
        LanguageAlias.GROOVY: 88,
        LanguageAlias.HASKELL: 61,
        LanguageAlias.JAVA: 62,
        LanguageAlias.JAVASCRIPT: 63,
        LanguageAlias.JAVA_OPENJDK: 62,
        LanguageAlias.KOTLIN: 78,
        LanguageAlias.LUA: 64,
        LanguageAlias.MULTI_FILE: 89,
        LanguageAlias.OBJECTIVE_C: 79,
        LanguageAlias.OCAML: 65,
        LanguageAlias.OCTAVE: 66,
        LanguageAlias.PASCAL: 67,
        LanguageAlias.PERL: 85,
        LanguageAlias.PHP: 68,
        LanguageAlias.PLAIN_TEXT: 43,
        LanguageAlias.PROLOG: 69,
        LanguageAlias.PYTHON: 71,
        LanguageAlias.PYTHON2: 70,
        LanguageAlias.PYTHON3: 71,
        LanguageAlias.R: 80,
        LanguageAlias.RUBY: 72,
        LanguageAlias.RUST: 73,
        LanguageAlias.SCALA: 81,
        LanguageAlias.SQLITE: 82,
        LanguageAlias.SWIFT: 83,
        LanguageAlias.TYPESCRIPT: 74,
        LanguageAlias.VISUAL_BASIC: 84,
    },
    "1.13.1-extra": {
        LanguageAlias.BOSQUE: 11,
        LanguageAlias.C: 1,
        LanguageAlias.C3: 3,
        LanguageAlias.CPP: 2,
        LanguageAlias.CPP_CLANG: 2,
        LanguageAlias.CPP_TEST: 12,
        LanguageAlias.CPP_TEST_CLANG: 15,
        LanguageAlias.CPP_TEST_GCC: 12,
        LanguageAlias.CSHARP: 22,
        LanguageAlias.CSHARP_MONO: 22,
        LanguageAlias.CSHARP_DOTNET: 21,
        LanguageAlias.CSHARP_TEST: 23,
        LanguageAlias.C_CLANG: 1,
        LanguageAlias.FSHARP: 24,
        LanguageAlias.JAVA: 4,
        LanguageAlias.JAVA_OPENJDK: 4,
        LanguageAlias.JAVA_TEST: 5,
        LanguageAlias.MPI_C: 6,
        LanguageAlias.MPI_CPP: 7,
        LanguageAlias.MPI_PYTHON: 8,
        LanguageAlias.MULTI_FILE: 89,
        LanguageAlias.NIM: 9,
        LanguageAlias.PYTHON: 10,
        LanguageAlias.PYTHON3: 10,
        LanguageAlias.PYTHON_FOR_ML: 10,
        LanguageAlias.VISUAL_BASIC: 20,
    },
    "1.14.0": {
        LanguageAlias.ASSEMBLY: 45,
        LanguageAlias.BASH: 46,
        LanguageAlias.BASIC: 47,
        LanguageAlias.C: 103,
        LanguageAlias.CLOJURE: 86,
        LanguageAlias.COBOL: 77,
        LanguageAlias.COMMON_LISP: 55,
        LanguageAlias.CPP: 105,
        LanguageAlias.CPP_CLANG: 76,
        LanguageAlias.CPP_GCC: 105,
        LanguageAlias.CSHARP: 51,
        LanguageAlias.CSHARP_MONO: 51,
        LanguageAlias.C_CLANG: 110,
        LanguageAlias.C_GCC: 103,
        LanguageAlias.D: 56,
        LanguageAlias.DART: 90,
        LanguageAlias.ELIXIR: 57,
        LanguageAlias.ERLANG: 58,
        LanguageAlias.EXECUTABLE: 44,
        LanguageAlias.FORTRAN: 59,
        LanguageAlias.FSHARP: 87,
        LanguageAlias.GO: 107,
        LanguageAlias.GROOVY: 88,
        LanguageAlias.HASKELL: 61,
        LanguageAlias.JAVA: 62,
        LanguageAlias.JAVAFX: 96,
        LanguageAlias.JAVASCRIPT: 102,
        LanguageAlias.JAVA_JDK: 91,
        LanguageAlias.JAVA_OPENJDK: 62,
        LanguageAlias.KOTLIN: 111,
        LanguageAlias.LUA: 64,
        LanguageAlias.MULTI_FILE: 89,
        LanguageAlias.OBJECTIVE_C: 79,
        LanguageAlias.OCAML: 65,
        LanguageAlias.OCTAVE: 66,
        LanguageAlias.PASCAL: 67,
        LanguageAlias.PERL: 85,
        LanguageAlias.PHP: 98,
        LanguageAlias.PLAIN_TEXT: 43,
        LanguageAlias.PROLOG: 69,
        LanguageAlias.PYTHON: 109,
        LanguageAlias.PYTHON2: 70,
        LanguageAlias.PYTHON3: 109,
        LanguageAlias.R: 99,
        LanguageAlias.RUBY: 72,
        LanguageAlias.RUST: 108,
        LanguageAlias.SCALA: 112,
        LanguageAlias.SQLITE: 82,
        LanguageAlias.SWIFT: 83,
        LanguageAlias.TYPESCRIPT: 101,
        LanguageAlias.VISUAL_BASIC: 84,
    },
    "1.14.0-extra": {
        LanguageAlias.BOSQUE: 11,
        LanguageAlias.C: 1,
        LanguageAlias.C3: 3,
        LanguageAlias.CPP: 2,
        LanguageAlias.CPP_CLANG: 2,
        LanguageAlias.CPP_TEST: 12,
        LanguageAlias.CPP_TEST_CLANG: 15,
        LanguageAlias.CPP_TEST_GCC: 12,
        LanguageAlias.CSHARP: 30,
        LanguageAlias.CSHARP_MONO: 22,
        LanguageAlias.CSHARP_DOTNET: 30,
        LanguageAlias.CSHARP_TEST: 23,
        LanguageAlias.C_CLANG: 1,
        LanguageAlias.FSHARP: 24,
        LanguageAlias.JAVA: 4,
        LanguageAlias.JAVA_OPENJDK: 4,
        LanguageAlias.JAVA_TEST: 5,
        LanguageAlias.MPI_C: 6,
        LanguageAlias.MPI_CPP: 7,
        LanguageAlias.MPI_PYTHON: 8,
        LanguageAlias.MULTI_FILE: 89,
        LanguageAlias.NIM: 9,
        LanguageAlias.PYTHON: 32,
        LanguageAlias.PYTHON2: 26,
        LanguageAlias.PYTHON2_PYPY: 26,
        LanguageAlias.PYTHON3: 32,
        LanguageAlias.PYTHON3_PYPY: 28,
        LanguageAlias.PYTHON_FOR_ML: 32,
        LanguageAlias.VISUAL_BASIC: 20,
    },
}



import copy
from datetime import datetime
from typing import Any, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator, UUID4

from .base_types import Iterable, LanguageAlias, Status
from .common import decode, encode
from .filesystem import Filesystem

ENCODED_REQUEST_FIELDS = {
    "source_code",
    "additional_files",
    "stdin",
    "expected_output",
}
ENCODED_RESPONSE_FIELDS = {
    "stdout",
    "stderr",
    "compile_output",
    # "post_execution_filesystem",
}
ENCODED_FIELDS = ENCODED_REQUEST_FIELDS | ENCODED_RESPONSE_FIELDS
EXTRA_REQUEST_FIELDS = {
    "compiler_options",
    "command_line_arguments",
    "cpu_time_limit",
    "cpu_extra_time",
    "wall_time_limit",
    "memory_limit",
    "stack_limit",
    "max_processes_and_or_threads",
    "enable_per_process_and_thread_time_limit",
    "enable_per_process_and_thread_memory_limit",
    "max_file_size",
    "redirect_stderr_to_stdout",
    "enable_network",
    "number_of_runs",
    "callback_url",
}
EXTRA_RESPONSE_FIELDS = {
    "message",
    "exit_code",
    "exit_signal",
    "status",
    "created_at",
    "finished_at",
    "token",
    "time",
    "wall_time",
    "memory",
}
REQUEST_FIELDS = ENCODED_REQUEST_FIELDS | EXTRA_REQUEST_FIELDS
RESPONSE_FIELDS = ENCODED_RESPONSE_FIELDS | EXTRA_RESPONSE_FIELDS
FIELDS = REQUEST_FIELDS | RESPONSE_FIELDS
SKIP_FIELDS = {"language_id", "language", "status_id"}
DATETIME_FIELDS = {"created_at", "finished_at"}
FLOATING_POINT_FIELDS = {
    "cpu_time_limit",
    "cpu_extra_time",
    "time",
    "wall_time",
    "wall_time_limit",
}

Submissions = Iterable["Submission"]


class Submission(BaseModel):
    """
    Stores a representation of a Submission to/from Judge0.

    Parameters
    ----------
    source_code : str, optional
        The source code to be executed.
    language : LanguageAlias or int, optional
        The programming language of the source code. Defaults to `LanguageAlias.PYTHON`.
    additional_files : base64 encoded string, optional
        Additional files that should be available alongside the source code.
        Value of this string should represent the content of a .zip that
        contains additional files. This attribute is required for multi-file
        programs.
    compiler_options : str, optional
        Options for the compiler (i.e. compiler flags).
    command_line_arguments : str, optional
        Command line arguments for the program.
    stdin : str, optional
        Input to be fed via standard input during execution.
    expected_output : str, optional
        The expected output of the program.
    cpu_time_limit : float, optional
        Maximum CPU time allowed for execution, in seconds. Time in which the
        OS assigns the processor to different tasks is not counted. Depends on
        configuration.
    cpu_extra_time : float, optional
        Additional CPU time allowance in case of time extension. Depends on
        configuration.
    wall_time_limit : float, optional
        Maximum wall clock time allowed for execution, in seconds. Depends on
        configuration.
    memory_limit : float, optional
        Maximum memory allocation allowed for the process, in kilobytes.
        Depends on configuration.
    stack_limit : int, optional
        Maximum stack size allowed, in kilobytes. Depends on configuration.
    max_processes_and_or_threads : int, optional
        Maximum number of processes and/or threads program can create. Depends
        on configuration.
    enable_per_process_and_thread_time_limit : bool, optional
        If True, enforces time limits per process/thread. Depends on
        configuration.
    enable_per_process_and_thread_memory_limit : bool, optional
        If True, enforces memory limits per process/thread. Depends on
        configuration.
    max_file_size : int, optional
        Maximum file size allowed for output files, in kilobytes. Depends on
        configuration.
    redirect_stderr_to_stdout : bool, optional
        If True, redirects standard error output to standard output.
    enable_network : bool, optional
        If True, enables network access during execution.
    number_of_runs : int, optional
        Number of times the code should be executed.
    callback_url : str, optional
        URL for a callback to report execution results or status.
    """

    source_code: Optional[Union[str, bytes]] = Field(default=None, repr=True)
    language: Union[LanguageAlias, int] = Field(
        default=LanguageAlias.PYTHON_FOR_ML,
        repr=True,
    )
    additional_files: Optional[Union[str, Filesystem]] = Field(default=None, repr=True)
    compiler_options: Optional[str] = Field(default=None, repr=True)
    command_line_arguments: Optional[str] = Field(default=None, repr=True)
    stdin: Optional[str] = Field(default=None, repr=True)
    expected_output: Optional[str] = Field(default=None, repr=True)
    cpu_time_limit: Optional[float] = Field(default=None, repr=True)
    cpu_extra_time: Optional[float] = Field(default=None, repr=True)
    wall_time_limit: Optional[float] = Field(default=None, repr=True)
    memory_limit: Optional[float] = Field(default=None, repr=True)
    stack_limit: Optional[int] = Field(default=None, repr=True)
    max_processes_and_or_threads: Optional[int] = Field(default=None, repr=True)
    enable_per_process_and_thread_time_limit: Optional[bool] = Field(
        default=None, repr=True
    )
    enable_per_process_and_thread_memory_limit: Optional[bool] = Field(
        default=None, repr=True
    )
    max_file_size: Optional[int] = Field(default=None, repr=True)
    redirect_stderr_to_stdout: Optional[bool] = Field(default=None, repr=True)
    enable_network: Optional[bool] = Field(default=None, repr=True)
    number_of_runs: Optional[int] = Field(default=None, repr=True)
    callback_url: Optional[str] = Field(default=None, repr=True)

    # Post-execution submission attributes.
    stdout: Optional[str] = Field(default=None, repr=True)
    stderr: Optional[str] = Field(default=None, repr=True)
    compile_output: Optional[str] = Field(default=None, repr=True)
    message: Optional[str] = Field(default=None, repr=True)
    exit_code: Optional[int] = Field(default=None, repr=True)
    exit_signal: Optional[int] = Field(default=None, repr=True)
    status: Optional[Status] = Field(default=None, repr=True)
    created_at: Optional[datetime] = Field(default=None, repr=True)
    finished_at: Optional[datetime] = Field(default=None, repr=True)
    token: Optional[UUID4] = Field(default=None, repr=True)
    time: Optional[float] = Field(default=None, repr=True)
    wall_time: Optional[float] = Field(default=None, repr=True)
    memory: Optional[float] = Field(default=None, repr=True)
    post_execution_filesystem: Optional[Filesystem] = Field(default=None, repr=True)

    model_config = ConfigDict(extra="ignore")

    @field_validator(*ENCODED_FIELDS, mode="before")
    @classmethod
    def process_encoded_fields(cls, value: str) -> Optional[str]:
        """Validate all encoded attributes."""
        if value is None:
            return None
        else:
            try:
                return decode(value)
            except Exception:
                return value

    @field_validator("post_execution_filesystem", mode="before")
    @classmethod
    def process_post_execution_filesystem(cls, content: str) -> Filesystem:
        """Validate post_execution_filesystem attribute."""
        return Filesystem(content=content)

    @field_validator("status", mode="before")
    @classmethod
    def process_status(cls, value: dict) -> Status:
        """Validate status attribute."""
        return Status(value["id"])

    @field_validator("language", mode="before")
    @classmethod
    def process_language(
        cls, value: Union[LanguageAlias, dict]
    ) -> Union[LanguageAlias, int]:
        """Validate status attribute."""
        if isinstance(value, dict):
            return value["id"]
        else:
            return value

    def set_attributes(self, attributes: dict[str, Any]) -> None:
        """Set submissions attributes while taking into account different
        attribute's types.

        Parameters
        ----------
        attributes : dict
            Key-value pairs of Submission attributes and the corresponding
            value.
        """
        for attr, value in attributes.items():
            if attr in SKIP_FIELDS:
                continue

            if attr in ENCODED_FIELDS:
                value = decode(value) if value else None
            elif attr == "status":
                value = Status(value["id"])
            elif attr in DATETIME_FIELDS and value is not None:
                value = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
            elif attr in FLOATING_POINT_FIELDS and value is not None:
                value = float(value)
            elif attr == "post_execution_filesystem":
                value = Filesystem(content=value)

            setattr(self, attr, value)

    def as_body(self, client: "Client") -> dict:
        """Prepare Submission as a dictionary while taking into account
        the client's restrictions.
        """
        body = {
            "source_code": encode(self.source_code),
            "language_id": client.get_language_id(self.language),
        }

        for field in ENCODED_REQUEST_FIELDS:
            value = getattr(self, field)
            if value is not None:
                body[field] = encode(value)

        for field in EXTRA_REQUEST_FIELDS:
            value = getattr(self, field)
            if value is not None:
                body[field] = value

        return body

    def is_done(self) -> bool:
        """Check if submission is finished processing.

        Submission is considered finished if the submission status is not
        IN_QUEUE and not PROCESSING.
        """
        if self.status is None:
            return False
        else:
            return self.status not in (Status.IN_QUEUE, Status.PROCESSING)

    def pre_execution_copy(self) -> "Submission":
        """Create a deep copy of a submission."""
        new_submission = Submission()
        for attr in REQUEST_FIELDS:
            setattr(new_submission, attr, copy.deepcopy(getattr(self, attr)))
        new_submission.language = self.language
        return new_submission

    def __iter__(self):
        if self.post_execution_filesystem is None:
            return iter([])
        else:
            return iter(self.post_execution_filesystem)



import copy
import io
import zipfile

from base64 import b64decode, b64encode
from typing import Optional, Union

from pydantic import BaseModel

from .base_types import Iterable


class File(BaseModel):
    """File object for storing file content.

    Parameters
    ----------
    name : str
        File name.
    content : str or bytes, optional
        File content. If str is provided, it will be encoded to bytes.
    """

    name: str
    content: Optional[Union[str, bytes]] = None

    def __init__(self, **data):
        super().__init__(**data)
        # Let's keep content attribute internally encoded as bytes.
        if isinstance(self.content, str):
            self.content = self.content.encode()
        elif isinstance(self.content, bytes):
            self.content = self.content
        else:
            self.content = b""

    def __str__(self):
        return self.content.decode(errors="backslashreplace")


class Filesystem(BaseModel):
    """Filesystem object for storing multiple files.

    Parameters
    ----------
    content : str or bytes or File or Iterable[File] or Filesystem, optional
        Filesystem content. If str or bytes is provided, it will be decoded to
        files.
    """

    files: list[File] = []

    def __init__(self, **data):
        content = data.pop("content", None)
        super().__init__(**data)
        self.files = []

        if isinstance(content, (bytes, str)):
            if isinstance(content, bytes):
                zip_bytes = content
            else:
                zip_bytes = b64decode(content.encode())

            with zipfile.ZipFile(io.BytesIO(zip_bytes), "r") as zip_file:
                for file_name in zip_file.namelist():
                    with zip_file.open(file_name) as fp:
                        self.files.append(File(name=file_name, content=fp.read()))
        elif isinstance(content, Iterable):
            self.files = list(content)
        elif isinstance(content, File):
            self.files = [content]
        elif isinstance(content, Filesystem):
            self.files = copy.deepcopy(content.files)
        elif content is None:
            self.files = []
        else:
            raise ValueError(
                "Unsupported type for content argument. Expected "
                "one of str, bytes, File, Iterable[File], or Filesystem, "
                f"got {type(content)}."
            )

    def __repr__(self) -> str:
        content_encoded = b64encode(self.encode()).decode()
        return f"{self.__class__.__name__}(content={content_encoded!r})"

    def encode(self) -> bytes:
        """Encode Filesystem object to bytes."""
        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "w") as zip_file:
            for file in self.files:
                zip_file.writestr(file.name, file.content)
        return zip_buffer.getvalue()

    def __str__(self) -> str:
        """Create string representation of Filesystem object."""
        return b64encode(self.encode()).decode()

    def __iter__(self):
        return iter(self.files)



import judge0
import pytest

from judge0 import Flavor, LanguageAlias, Submission
from judge0.api import _resolve_client

DEFAULT_CLIENTS = (
    "atd_ce_client",
    "atd_extra_ce_client",
    "rapid_ce_client",
    "rapid_extra_ce_client",
    "judge0_cloud_ce_client",
    "judge0_cloud_extra_ce_client",
)


@pytest.mark.parametrize("client", DEFAULT_CLIENTS)
def test_resolve_client_with_explicit_client(client, request):
    client = request.getfixturevalue(client)
    assert _resolve_client(client) is client


@pytest.mark.parametrize(
    "flavor,expected_client",
    [
        [
            Flavor.CE,
            "JUDGE0_IMPLICIT_CE_CLIENT",
        ],
        [
            Flavor.EXTRA_CE,
            "JUDGE0_IMPLICIT_EXTRA_CE_CLIENT",
        ],
    ],
)
def test_resolve_client_with_flavor(
    flavor,
    expected_client,
):
    # We have to use getattr since both implicit clients are initially None.
    assert _resolve_client(client=flavor) is getattr(judge0, expected_client)


@pytest.mark.parametrize(
    "submissions",
    [
        [],
        None,
    ],
)
def test_resolve_client_empty_submissions_argument(submissions):
    with pytest.raises(ValueError):
        _resolve_client(submissions=submissions)


def test_resolve_client_no_common_client_for_submissions():
    cpp_submission = Submission(
        source_code="",  # source code is not important in this test
        language=LanguageAlias.CPP_GCC,
    )

    py_submission = Submission(
        source_code="",  # source code is not important in this test
        language=LanguageAlias.PYTHON_FOR_ML,
    )

    submissions = [cpp_submission, py_submission]

    with pytest.raises(RuntimeError):
        _resolve_client(submissions=submissions)


def test_resolve_client_common_ce_client():
    cpp_submission = Submission(
        source_code="",  # source code is not important in this test
        language=LanguageAlias.CPP_GCC,
    )

    py_submission = Submission(
        source_code="",  # source code is not important in this test
        language=LanguageAlias.PYTHON,
    )

    submissions = [cpp_submission, py_submission]

    assert _resolve_client(submissions=submissions) is judge0.JUDGE0_IMPLICIT_CE_CLIENT


def test_resolve_client_common_extra_ce_client():
    cpp_submission = Submission(
        source_code="",  # source code is not important in this test
        language=LanguageAlias.CPP_CLANG,
    )

    py_submission = Submission(
        source_code="",  # source code is not important in this test
        language=LanguageAlias.PYTHON_FOR_ML,
    )

    submissions = [cpp_submission, py_submission]

    assert (
        _resolve_client(submissions=submissions)
        is judge0.JUDGE0_IMPLICIT_EXTRA_CE_CLIENT
    )



from base64 import b64decode, b64encode
from itertools import islice
from typing import Union

from judge0.base_types import Encodable


def encode(content: Union[bytes, str, Encodable]) -> str:
    """Encode content to base64 string."""
    if isinstance(content, bytes):
        return b64encode(content).decode()
    if isinstance(content, str):
        return b64encode(content.encode()).decode()
    if isinstance(content, Encodable):
        return b64encode(content.encode()).decode()
    raise ValueError(f"Unsupported type. Expected bytes or str, got {type(content)}!")


def decode(content: Union[bytes, str]) -> str:
    """Decode base64 encoded content."""
    if isinstance(content, bytes):
        return b64decode(
            content.decode(errors="backslashreplace"), validate=True
        ).decode(errors="backslashreplace")
    if isinstance(content, str):
        return b64decode(content.encode(), validate=True).decode(
            errors="backslashreplace"
        )
    raise ValueError(f"Unsupported type. Expected bytes or str, got {type(content)}!")


def batched(iterable, n):
    """Iterate over an iterable in batches of a specified size.

    Adapted from https://docs.python.org/3/library/itertools.html#itertools.batched.
    """
    if n < 1:
        raise ValueError("n must be at least one")
    iterator = iter(iterable)
    while True:
        batch = tuple(islice(iterator, n))
        if not batch:
            break
        yield batch



"""Module containing different utility functions for Judge0 Python SDK."""

from functools import wraps
from http import HTTPStatus

from requests import HTTPError

from .errors import PreviewClientLimitError


def is_http_too_many_requests_error(exception: Exception) -> bool:
    return (
        isinstance(exception, HTTPError)
        and exception.response is not None
        and exception.response.status_code == HTTPStatus.TOO_MANY_REQUESTS
    )


def handle_too_many_requests_error_for_preview_client(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HTTPError as err:
            if is_http_too_many_requests_error(exception=err):
                # If the raised exception is inside the one of the Judge0 Cloud clients
                # let's check if we are dealing with the implicit client.
                if args:
                    instance = args[0]
                    class_name = instance.__class__.__name__
                    # Check if we are using a preview version of the client.
                    if (
                        class_name in ("Judge0CloudCE", "Judge0CloudExtraCE")
                        and instance.api_key is None
                    ):
                        raise PreviewClientLimitError(
                            "You are using a preview version of a client and "
                            f"you've hit a rate limit on it. Visit {instance.HOME_URL} "
                            "to get your authentication credentials."
                        ) from err
                else:
                    raise err from None
            else:
                raise err from None
        except Exception as err:
            raise err from None

    return wrapper



import pytest

DEFAULT_CLIENTS = (
    "atd_ce_client",
    "atd_extra_ce_client",
    "rapid_ce_client",
    "rapid_extra_ce_client",
    "judge0_cloud_ce_client",
    "judge0_cloud_extra_ce_client",
)


@pytest.mark.parametrize("client", DEFAULT_CLIENTS)
def test_get_about(client, request):
    client = request.getfixturevalue(client)
    client.get_about()


@pytest.mark.parametrize("client", DEFAULT_CLIENTS)
def test_get_config_info(client, request):
    client = request.getfixturevalue(client)
    client.get_config_info()


@pytest.mark.parametrize("client", DEFAULT_CLIENTS)
def test_get_languages(client, request):
    client = request.getfixturevalue(client)
    client.get_languages()


@pytest.mark.parametrize("client", DEFAULT_CLIENTS)
def test_get_statuses(client, request):
    client = request.getfixturevalue(client)
    client.get_statuses()


@pytest.mark.parametrize("client", DEFAULT_CLIENTS)
def test_is_language_supported_multi_file_submission(client, request):
    client = request.getfixturevalue(client)
    assert client.is_language_supported(89)


@pytest.mark.parametrize("client", DEFAULT_CLIENTS)
def test_is_language_supported_non_valid_lang_id(client, request):
    client = request.getfixturevalue(client)
    assert not client.is_language_supported(-1)



from typing import Optional, Union

from .base_types import Flavor, Iterable, TestCase, TestCases, TestCaseType
from .clients import Client
from .common import batched
from .errors import ClientResolutionError
from .retry import RegularPeriodRetry, RetryStrategy
from .submission import Submission, Submissions


def get_client(flavor: Flavor = Flavor.CE) -> Client:
    """Resolve client from API keys from environment or default to preview client.

    Parameters
    ----------
    flavor : Flavor
        Flavor of Judge0 Client.

    Returns
    -------
    Client
        An object of base type Client and the specified flavor.
    """
    from . import _get_implicit_client

    if isinstance(flavor, Flavor):
        return _get_implicit_client(flavor=flavor)
    else:
        raise ValueError(
            "Expected argument flavor to be of of type enum Flavor, "
            f"got {type(flavor)}."
        )


def _resolve_client(
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
) -> Client:
    """Resolve a client from flavor or submission(s) arguments.

    Parameters
    ----------
    client : Client or Flavor, optional
        A Client object or flavor of client. Returns the client if not None.
    submissions: Submission or Submissions, optional
        Submission(s) used to determine the suitable client.

    Returns
    -------
    Client
        An object of base type Client.

    Raises
    ------
    ClientResolutionError
        If there is no implemented client that supports all the languages specified
        in the submissions.
    """
    # User explicitly passed a client.
    if isinstance(client, Client):
        return client

    # NOTE: At the moment, we do not support the option to check if explicit
    # flavor of a client supports the submissions, i.e. submissions argument is
    # ignored if flavor argument is provided.

    if isinstance(client, Flavor):
        return get_client(client)

    if client is None:
        if (
            isinstance(submissions, Iterable) and len(submissions) == 0
        ) or submissions is None:
            raise ValueError("Client cannot be determined from empty submissions.")

    # client is None and we have to determine a flavor of the client from the
    # the submission's languages.
    if isinstance(submissions, Submission):
        submissions = [submissions]

    # Check which client supports all languages from the provided submissions.
    languages = [submission.language for submission in submissions]

    for flavor in Flavor:
        client = get_client(flavor)
        if client is not None and all(
            (client.is_language_supported(lang) for lang in languages)
        ):
            return client

    raise ClientResolutionError(
        "Failed to resolve the client from submissions argument. "
        "None of the implicit clients supports all languages from the submissions. "
        "Please explicitly provide the client argument."
    )


def create_submissions(
    *,
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
) -> Union[Submission, Submissions]:
    """Universal function for creating submissions to the client.

    Parameters
    ----------
    client : Client or Flavor, optional
        A client or client flavor where submissions should be created.
    submissions: Submission or Submissions, optional
        Submission(s) to create.

    Raises
    ------
    ClientResolutionError
        Raised if client resolution fails.
    """
    client = _resolve_client(client=client, submissions=submissions)

    if isinstance(submissions, Submission):
        return client.create_submission(submissions)

    result_submissions = []
    for submission_batch in batched(
        submissions, client.config.max_submission_batch_size
    ):
        if len(submission_batch) > 1:
            result_submissions.extend(client.create_submissions(submission_batch))
        else:
            result_submissions.append(client.create_submission(submission_batch[0]))

    return result_submissions


def get_submissions(
    *,
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
    fields: Optional[Union[str, Iterable[str]]] = None,
) -> Union[Submission, Submissions]:
    """Get submission (status) from a client.

    Parameters
    ----------
    client : Client or Flavor, optional
        A client or client flavor where submissions should be checked.
    submissions : Submission or Submissions, optional
        Submission(s) to update.
    fields : str or sequence of str, optional
        Submission attributes that need to be updated. Defaults to all attributes.

    Raises
    ------
    ClientResolutionError
        Raised if client resolution fails.
    """
    client = _resolve_client(client=client, submissions=submissions)

    if isinstance(submissions, Submission):
        return client.get_submission(submissions, fields=fields)

    result_submissions = []
    for submission_batch in batched(
        submissions, client.config.max_submission_batch_size
    ):
        if len(submission_batch) > 1:
            result_submissions.extend(
                client.get_submissions(submission_batch, fields=fields)
            )
        else:
            result_submissions.append(
                client.get_submission(submission_batch[0], fields=fields)
            )

    return result_submissions


def wait(
    *,
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
    retry_strategy: Optional[RetryStrategy] = None,
) -> Union[Submission, Submissions]:
    """Wait for all the submissions to finish.

    Parameters
    ----------
    client : Client or Flavor, optional
        A client or client flavor where submissions should be checked.
    submissions : Submission or Submissions
        Submission(s) to wait for.
    retry_strategy : RetryStrategy, optional
        A retry strategy.

    Returns
    -------
    Submission or Submissions
        A single submission or a list of submissions.

    Raises
    ------
    ClientResolutionError
        Raised if client resolution fails.
    """
    client = _resolve_client(client, submissions)

    if retry_strategy is None:
        if client.retry_strategy is None:
            retry_strategy = RegularPeriodRetry()
        else:
            retry_strategy = client.retry_strategy

    if isinstance(submissions, Submission):
        submissions_list = [submissions]
    else:
        submissions_list = submissions

    submissions_to_check = {
        submission.token: submission for submission in submissions_list
    }

    while len(submissions_to_check) > 0 and not retry_strategy.is_done():
        get_submissions(client=client, submissions=list(submissions_to_check.values()))
        finished_submissions = [
            token
            for token, submission in submissions_to_check.items()
            if submission.is_done()
        ]
        for token in finished_submissions:
            submissions_to_check.pop(token)

        # Don't wait if there is no submissions to check for anymore.
        if len(submissions_to_check) == 0:
            break

        retry_strategy.wait()
        retry_strategy.step()

    return submissions


def create_submissions_from_test_cases(
    submissions: Union[Submission, Submissions],
    test_cases: Optional[Union[TestCaseType, TestCases]] = None,
) -> Union[Submission, list[Submission]]:
    """Create submissions from the submission and test case pairs.

    Function always returns a deep copy so make sure you are using the
    returned submission(s).

    Parameters
    ----------
    submissions : Submission or Submissions
        Base submission(s) that need to be expanded with test cases.
    test_cases: TestCaseType or TestCases
        Test cases.

    Returns
    -------
    Submissions or Submissions
        A single submission if submissions arguments is of type Submission or
        source_code argument is provided, and test_cases argument is of type
        TestCase. Otherwise returns a list of submissions.
    """
    if isinstance(submissions, Submission):
        submissions_list = [submissions]
    else:
        submissions_list = submissions

    if isinstance(test_cases, TestCase) or test_cases is None:
        test_cases_list = [test_cases]
        multiple_test_cases = False
    else:
        try:
            # Let's assume that we are dealing with multiple test_cases that
            # can be created from test_cases argument. If this fails, i.e.
            # raises a ValueError, we know we are dealing with a test_cases=dict,
            # or test_cases=["in", "out"], or test_cases=tuple("in", "out").
            test_cases_list = [TestCase.from_record(tc) for tc in test_cases]

            # It is possible to send test_cases={}, or test_cases=[], or
            # test_cases=tuple([]). In this case, we are treating that as None.
            if len(test_cases) > 0:
                multiple_test_cases = True
            else:
                multiple_test_cases = False
                test_cases_list = [None]
        except ValueError:
            test_cases_list = [test_cases]
            multiple_test_cases = False

    test_cases_list = [TestCase.from_record(test_case=tc) for tc in test_cases_list]

    all_submissions = []
    for submission in submissions_list:
        for test_case in test_cases_list:
            submission_copy = submission.pre_execution_copy()
            if test_case is not None:
                submission_copy.stdin = test_case.input
                submission_copy.expected_output = test_case.expected_output
            all_submissions.append(submission_copy)

    if isinstance(submissions, Submission) and (not multiple_test_cases):
        return all_submissions[0]
    else:
        return all_submissions


def _execute(
    *,
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
    source_code: Optional[str] = None,
    test_cases: Optional[Union[TestCaseType, TestCases]] = None,
    wait_for_result: bool = False,
    **kwargs,
) -> Union[Submission, Submissions]:

    if submissions is not None and source_code is not None:
        raise ValueError(
            "Both submissions and source_code arguments are provided. "
            "Provide only one of the two."
        )
    if submissions is None and source_code is None:
        raise ValueError("Neither source_code nor submissions argument are provided.")

    # Internally, let's rely on Submission's dataclass.
    if source_code is not None:
        submissions = Submission(source_code=source_code, **kwargs)

    client = _resolve_client(client=client, submissions=submissions)
    all_submissions = create_submissions_from_test_cases(submissions, test_cases)
    all_submissions = create_submissions(client=client, submissions=all_submissions)

    if wait_for_result:
        return wait(client=client, submissions=all_submissions)
    else:
        return all_submissions


def async_execute(
    *,
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
    source_code: Optional[str] = None,
    test_cases: Optional[Union[TestCaseType, TestCases]] = None,
    **kwargs,
) -> Union[Submission, Submissions]:
    """Create submission(s).

    Aliases: `async_run`.

    Parameters
    ----------
    client : Client or Flavor, optional
        A client where submissions should be created. If None, will try to be
        resolved.
    submissions : Submission or Submissions, optional
        Submission or submissions for execution.
    source_code : str, optional
        A source code of a program.
    test_cases : TestCaseType or TestCases, optional
        A single test or a list of test cases
    **kwargs : dict
        Additional keyword arguments to pass to the Submission constructor.

    Returns
    -------
    Submission or Submissions
        A single submission if submissions arguments is of type Submission or
        source_code argument is provided, and test_cases argument is of type
        TestCase. Otherwise returns a list of submissions.

    Raises
    ------
    ClientResolutionError
        If client cannot be resolved from the submissions or the flavor.
    ValueError
        If both or neither submissions and source_code arguments are provided.
    """
    return _execute(
        client=client,
        submissions=submissions,
        source_code=source_code,
        test_cases=test_cases,
        wait_for_result=False,
        **kwargs,
    )


def sync_execute(
    *,
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
    source_code: Optional[str] = None,
    test_cases: Optional[Union[TestCaseType, TestCases]] = None,
    **kwargs,
) -> Union[Submission, Submissions]:
    """Create submission(s) and wait for their finish.

    Aliases: `execute`, `run`, `sync_run`.

    Parameters
    ----------
    client : Client or Flavor, optional
        A client where submissions should be created. If None, will try to be
        resolved.
    submissions : Submission or Submissions, optional
        Submission(s) for execution.
    source_code: str, optional
        A source code of a program.
    test_cases: TestCaseType or TestCases, optional
        A single test or a list of test cases
    **kwargs : dict
        Additional keyword arguments to pass to the Submission constructor.

    Returns
    -------
    Submission or Submissions
        A single submission if submissions arguments is of type Submission or
        source_code argument is provided, and test_cases argument is of type
        TestCase. Otherwise returns a list of submissions.

    Raises
    ------
    ClientResolutionError
        If client cannot be resolved from the submissions or the flavor.
    ValueError
        If both or neither submissions and source_code arguments are provided.
    """
    return _execute(
        client=client,
        submissions=submissions,
        source_code=source_code,
        wait_for_result=True,
        test_cases=test_cases,
        **kwargs,
    )


execute = sync_execute
run = sync_execute
sync_run = sync_execute
async_run = async_execute



"""Separate file containing tests related to test case functionality."""

import judge0
import pytest
from judge0.api import create_submissions_from_test_cases
from judge0.base_types import LanguageAlias, Status, TestCase
from judge0.submission import Submission


@pytest.mark.parametrize(
    "test_case,expected_output",
    [
        [
            TestCase(input="input_1", expected_output="output_1"),
            TestCase(input="input_1", expected_output="output_1"),
        ],
        [
            tuple([]),
            TestCase(input=None, expected_output=None),
        ],
        [
            ("input_tuple",),
            TestCase(input="input_tuple", expected_output=None),
        ],
        [
            ("input_tuple", "output_tuple"),
            TestCase(input="input_tuple", expected_output="output_tuple"),
        ],
        [
            [],
            TestCase(input=None, expected_output=None),
        ],
        [
            ["input_list"],
            TestCase(input="input_list", expected_output=None),
        ],
        [
            ["input_list", "output_list"],
            TestCase(input="input_list", expected_output="output_list"),
        ],
        [
            {"input": "input_dict", "expected_output": "output_dict"},
            TestCase(input="input_dict", expected_output="output_dict"),
        ],
        [
            None,
            None,
        ],
    ],
)
def test_test_case_from_record(test_case, expected_output):
    assert TestCase.from_record(test_case) == expected_output


@pytest.mark.parametrize(
    "submissions,test_cases,expected_type",
    [
        [Submission(source_code=""), TestCase(), Submission],
        [[Submission(source_code="")], TestCase(), list],
        [Submission(source_code=""), [TestCase()], list],
        [[Submission(source_code="")], [TestCase()], list],
    ],
)
def test_create_submissions_from_test_cases_return_type(
    submissions, test_cases, expected_type
):
    output = create_submissions_from_test_cases(submissions, test_cases)
    assert type(output) is expected_type


class TestCreateSubmissionsFromTestCases:
    @pytest.mark.parametrize(
        "test_case,stdin,expected_output",
        [
            [TestCase(), None, None],
            [[], None, None],
            [{}, None, None],
            [tuple([]), None, None],
        ],
    )
    def test_empty_test_case(self, test_case, stdin, expected_output):
        submission = create_submissions_from_test_cases(
            Submission(), test_cases=test_case
        )

        assert (
            submission.stdin == stdin and submission.expected_output == expected_output
        )

    @pytest.mark.parametrize(
        "test_case,stdin,expected_output",
        [
            [TestCase(), None, None],
            [TestCase(input="input"), "input", None],
            [TestCase(expected_output="output"), None, "output"],
            [["input_list"], "input_list", None],
            [["input_list", "output_list"], "input_list", "output_list"],
            [{"input": "input_dict"}, "input_dict", None],
            [
                {"input": "input_dict", "expected_output": "output_dict"},
                "input_dict",
                "output_dict",
            ],
            [("input_tuple",), "input_tuple", None],
            [("input_tuple", "output_tuple"), "input_tuple", "output_tuple"],
        ],
    )
    def test_single_test_case(self, test_case, stdin, expected_output):
        submission = create_submissions_from_test_cases(
            Submission(), test_cases=test_case
        )

        assert (
            submission.stdin == stdin and submission.expected_output == expected_output
        )

    @pytest.mark.parametrize(
        "test_cases,stdin,expected_output",
        [
            [[TestCase()], None, None],
            [[TestCase(input="input")], "input", None],
            [[TestCase(expected_output="output")], None, "output"],
            [(["input_list"],), "input_list", None],
            [(["input_list", "output_list"],), "input_list", "output_list"],
            [({"input": "input_dict"},), "input_dict", None],
            [
                ({"input": "input_dict", "expected_output": "output_dict"},),
                "input_dict",
                "output_dict",
            ],
            [
                [
                    ("input_tuple",),
                ],
                "input_tuple",
                None,
            ],
            [
                [
                    ("input_tuple", "output_tuple"),
                ],
                "input_tuple",
                "output_tuple",
            ],
        ],
    )
    def test_single_test_case_in_iterable(self, test_cases, stdin, expected_output):
        submissions = create_submissions_from_test_cases(
            Submission(), test_cases=test_cases
        )

        for submission in submissions:
            assert (
                submission.stdin == stdin
                and submission.expected_output == expected_output
            )


@pytest.mark.parametrize(
    "source_code_or_submissions,test_cases,expected_status",
    [
        [
            "print(f'Hello, {input()}')",
            [TestCase("Judge0", "Hello, Judge0")],
            [Status.ACCEPTED],
        ],
        [
            "print(f'Hello, {input()}')",
            [
                TestCase("Judge0", "Hello, Judge0"),
                TestCase("pytest", "Hello, pytest"),
            ],
            [Status.ACCEPTED, Status.ACCEPTED],
        ],
        [
            Submission(
                source_code="print(f'Hello, {input()}')",
                language=LanguageAlias.PYTHON,
            ),
            [
                TestCase("Judge0", "Hello, Judge0"),
            ],
            [Status.ACCEPTED],
        ],
        [
            Submission(
                source_code="print(f'Hello, {input()}')",
                language=LanguageAlias.PYTHON,
            ),
            [
                TestCase("Judge0", "Hello, Judge0"),
                TestCase("pytest", "Hi, pytest"),
            ],
            [Status.ACCEPTED, Status.WRONG_ANSWER],
        ],
        [
            [
                Submission(
                    source_code="print(f'Hello, {input()}')",
                    language=LanguageAlias.PYTHON,
                ),
                Submission(
                    source_code="print(f'Hello,  {input()}')",
                    language=LanguageAlias.PYTHON,
                ),
            ],
            [
                TestCase("Judge0", "Hello, Judge0"),
                TestCase("pytest", "Hello, pytest"),
            ],
            [
                Status.ACCEPTED,
                Status.ACCEPTED,
                Status.WRONG_ANSWER,
                Status.WRONG_ANSWER,
            ],
        ],
    ],
)
def test_test_cases_from_run(
    source_code_or_submissions, test_cases, expected_status, request
):
    client = request.getfixturevalue("ce_client")

    if isinstance(source_code_or_submissions, str):
        submissions = judge0.run(
            client=client,
            source_code=source_code_or_submissions,
            test_cases=test_cases,
            language=LanguageAlias.PYTHON,
        )
    else:
        submissions = judge0.run(
            client=client,
            submissions=source_code_or_submissions,
            test_cases=test_cases,
        )

    assert [submission.status for submission in submissions] == expected_status


@pytest.mark.parametrize(
    "submissions,expected_status",
    [
        [
            Submission(
                source_code="print(f'Hello, {input()}')",
                language=LanguageAlias.PYTHON,
                stdin="Judge0",
                expected_output="Hello, Judge0",
            ),
            Status.ACCEPTED,
        ],
        [
            [
                Submission(
                    source_code="print(f'Hello, {input()}')",
                    language=LanguageAlias.PYTHON,
                    stdin="Judge0",
                    expected_output="Hello, Judge0",
                ),
                Submission(
                    source_code="print(f'Hello, {input()}')",
                    language=LanguageAlias.PYTHON,
                    stdin="pytest",
                    expected_output="Hello, pytest",
                ),
            ],
            [Status.ACCEPTED, Status.ACCEPTED],
        ],
    ],
)
def test_no_test_cases(submissions, expected_status, request):
    client = request.getfixturevalue("ce_client")

    submissions = judge0.run(
        client=client,
        submissions=submissions,
    )

    if isinstance(submissions, list):
        assert [submission.status for submission in submissions] == expected_status
    else:
        assert submissions.status == expected_status


@pytest.mark.parametrize("n_submissions", [42, 84])
def test_batched_test_cases(n_submissions, request):
    client = request.getfixturevalue("ce_client")
    submissions = [
        Submission(
            source_code=f"print({i})",
            language=LanguageAlias.PYTHON,
            expected_output=f"{i}",
        )
        for i in range(n_submissions)
    ]

    results = judge0.run(client=client, submissions=submissions)

    assert len(results) == n_submissions
    assert all([result.status == Status.ACCEPTED for result in results])
